from .annotations import *
from .base import EndPoint, Path
from ..definitions.descriptors import *
from ..definitions.metaclass import Array
from ..definitions.types import *

__all__ = ['GETAccounts', 'GETAccountID', 'GETAccountIDSummary', 'GETAccountIDInstruments',
           'PATCHAccountIDConfiguration', 'GETAccountIDChanges']


class GETAccounts(EndPoint):
    # the HTTP verb to use for this endpoint
    method = 'GET'

    # path to endpoint
    path = Path('/v3/accounts')

    # description of endpoint
    description = 'Get a list of all Accounts authorized for the provided token.'

    # parameters required to send to endpoint
    parameters = [
        {'name': 'Authorization', 'located': 'header', 'type': Authorization, 'description': 'str'},
    ]

    # valid responses
    responses = {200: {'accounts': Array(AccountProperties)}}

    # error msgs'
    error = (401, 405)


class GETAccountID(EndPoint):
    # the HTTP verb to use for this endpoint
    method = 'GET'

    # path to endpoint
    path = Path('/v3/accounts/', AccountID)

    # description of endpoint
    description = 'Get the full details for a single Account that a client has access to. ' \
                  'Full pending Order, open Trade and open Position representations are provided.'

    # parameters required to send to endpoint
    parameters = [
        {'name': 'Authorization', 'located': 'header', 'type': Authorization, 'description': 'str'},
        {'name': 'Accept-Datetime-Format', 'located': 'header', 'type': AcceptDatetimeFormat,
         'description': 'AcceptDatetimeFormat'},
        {'name': 'accountID', 'located': 'path', 'type': AccountID, 'description': 'AccountID'},
    ]

    # valid responses
    responses = {200: {'account': Account, 'lastTransactionID': TransactionID}}

    # error msgs'
    error = (400, 401, 405)


class GETAccountIDSummary(EndPoint):
    # the HTTP verb to use for this endpoint
    method = 'GET'

    # path to endpoint
    path = Path('/v3/accounts/', AccountID, '/summary')

    # description of endpoint
    description = 'Get a summary for a single Account that a client has access to.'

    # parameters required to send to endpoint
    parameters = [
        {'name': 'Authorization', 'located': 'header', 'type': Authorization, 'description': 'str'},
        {'name': 'Accept-Datetime-Format', 'located': 'header', 'type': AcceptDatetimeFormat,
         'description': 'AcceptDatetimeFormat'},
        {'name': 'accountID', 'located': 'path', 'type': AccountID, 'description': 'AccountID'},
    ]

    # valid responses
    responses = {200: {'account': AccountSummary, 'lastTransactionID': TransactionID}}

    # error msgs'
    error = (400, 401, 405)


class GETAccountIDInstruments(EndPoint):
    # the HTTP verb to use for this endpoint
    method = 'GET'

    # path to endpoint
    path = Path('/v3/accounts/', AccountID, '/instruments')

    # description of endpoint
    description = 'Get the list of tradeable instruments for the given Account.' \
                  'The list of tradeable instruments is dependent on the regulatory ' \
                  'division that the Account is located in, thus should be the same for all ' \
                  'Accounts owned by a single user.'

    # parameters required to send to endpoint
    parameters = [
        {'name': 'Authorization', 'located': 'header', 'type': Authorization, 'description': 'str'},
        {'name': 'accountID', 'located': 'path', 'type': AccountID, 'description': 'AccountID'},
        {'name': 'instruments', 'located': 'query', 'type': Instruments,
         'description': 'List of InstrumentName (csv)'},
    ]

    # valid responses
    responses = {200: {'instruments': Array(Instrument), 'lastTransactionID': TransactionID}}

    # error msgs'
    error = (400, 401, 405)


class PATCHAccountIDConfiguration(EndPoint):
    # the HTTP verb to use for this endpoint
    method = 'PATCH'

    # path to endpoint
    path = Path('/v3/accounts/', AccountID, '/configuration')

    # description of endpoint
    description = 'Set the client-configurable portions of an Account.'

    # parameters required to send to endpoint
    parameters = [
        {'name': 'Authorization', 'located': 'header', 'type': Authorization, 'description': 'str'},
        {'name': 'Accept-Datetime-Format', 'located': 'header', 'type': AcceptDatetimeFormat,
         'description': 'AcceptDatetimeFormat'},
        {'name': 'accountID', 'located': 'path', 'type': AccountID, 'description': 'AccountID'},
    ]

    # valid responses
    responses = {200: {'clientConfigureTransaction': ClientConfigureTransaction, 'lastTransactionID': TransactionID},
                 400: {'clientConfigureRejectTransaction': ClientConfigureRejectTransaction,
                       'lastTransactionID': TransactionID, 'errorCode': str, 'errorMessage': str},
                 403: {'clientConfigureRejectTransaction': ClientConfigureRejectTransaction,
                       'lastTransactionID': TransactionID, 'errorCode': str, 'errorMessage': str}}

    # error msgs'
    error = (401, 404, 405)

    # json schema representation
    request_schema = {'alias': Alias, 'marginRate': DecimalNumber}


class GETAccountIDChanges(EndPoint):
    # the HTTP verb to use for this endpoint
    method = 'GET'

    # path to endpoint
    path = Path('/v3/accounts/', AccountID, '/changes')
    # description of endpoint
    description = 'Endpoint used to poll an Account for its current state and changes since a specified TransactionID.'

    # parameters required to send to endpoint
    parameters = [
        {'name': 'Authorization', 'located': 'header', 'type': Authorization, 'description': 'str'},
        {'name': 'Accept-Datetime-Format', 'located': 'header', 'type': AcceptDatetimeFormat,
         'description': 'AcceptDatetimeFormat'},
        {'name': 'accountID', 'located': 'path', 'type': AccountID, 'description': 'AccountID'},
        {'name': 'sinceTransactionID', 'located': 'query', 'type': LastTransactionID, 'description': 'TransactionID'},
    ]

    # valid responses
    responses = {200: {'changes': AccountChanges, 'state': AccountChangesState, 'lastTransactionID': TransactionID}}

    # error msgs'
    error = (401, 404, 405, 416)



from .account import *
from .instrument import *
from .order import *
from .position import *
from .pricing import *
from .trade import *
from .transaction import *
from .user import *

__all__ = (
    account.__all__ +
    instrument.__all__ +
    order.__all__ +
    position.__all__ +
    pricing.__all__ +
    trade.__all__ +
    transaction.__all__ +
    user.__all__
)


import ujson as json

from .response import Response
from ..definitions.helpers import create_attribute
from ..endpoints.annotations import LastTransactionID
from ..endpoints.other_responses import other_responses


def _create_objects(schema, key, data):
    typ = schema.get(key)
    return key, create_attribute(typ, data)

async def _create_response(json_body, schema):
    return Response([_create_objects(schema, json_object, object_field)
                 for json_object, object_field in json_body.items()])

async def _rest_response(self, response, endpoint):
    async with response as resp:
        status = resp.status
        headers = resp.raw_headers
        json_body = await resp.json()

    print(f'RESPONSE STATUS: {status}')

    # Update client headers. Such as lastTransactionID and the like
    self.default_parameters.update(headers)
    last_transaction_id = json_body.get('lastTransactionID', None)
    if last_transaction_id:
        self.default_parameters.update({LastTransactionID: last_transaction_id})

    try:
        schema = endpoint.responses[status]  # look up the template to process the data
    except KeyError:
        schema = other_responses[status]  # See if a response status is an error code

    result = await _create_response(json_body, schema)
    result.raw_body = json_body
    return result


async def _stream_parser(response, endpoint):
    async with response as resp:
        schema = endpoint.responses[resp.status]
        async for line in resp.content:
            body = json.loads(line)
            key = body.get('type')
            json_body = {key: body}
            yield await _create_response(json_body, schema)


async def parse_response(self, response, endpoint):
    if endpoint.host == 'REST':
        result = await _rest_response(self, response, endpoint)
    else:
        result = _stream_parser(response, endpoint)
    return result



class LengthError(Exception):
    def __init__(self, output):
        super().__init__(output)


class IncorrectValue(Exception):
    def __init__(self, output):
        super().__init__(output)


class DescriptorProtocol(object):
    value = None

    def __init__(self, name):
        self.name = name

    def __set__(self, instance, value):
        setattr(instance, self.name, value)

    def __get__(self, instance, owner):
        if instance is None:
            return self
        return getattr(instance, self.name)

    def __delete__(self, instance, value):
        delattr(instance, self.name)


class Descriptor(DescriptorProtocol):
    typ = None
    example = None
    values = None

    def __new__(cls, value=None, *, name=None):
        if value is not None:
            return cls.typ(value)
        else:
            return super().__new__(cls)

    def type_check(self, value):
        typ = getattr(self, 'typ', None)
        if typ:
            try:
                value = typ(value)
            except ValueError:
                msg = f'{self.name} must be of type {self.typ.__name__}. ' \
                      f'Or {typ.__name__}({value}) returns {typ}. ' \
                      f'Was {type(value)}'
                raise TypeError(msg)
            else:
                return value

    def __set__(self, instance, value):

        value = self.type_check(value)

        if getattr(self, 'example', None):
            try:
                assert len(self.example) == len(value)
            except AssertionError:
                msg = f'{value} does not match length of example {self.example}. Attribute {type(self)}'
                raise LengthError(msg)

        if getattr(self, 'values', None):
            try:
                assert value in self.values
            except AssertionError:
                possible_values = ', '.join(self.values)
                msg = f'{value} must be in {self.values}. Possible values are {possible_values}'
                raise IncorrectValue(msg)

        setattr(instance, self.name, value)



import asyncio

from async_v20 import OandaClient


async def stream(instruments):
    async with OandaClient() as client:
        async for data in await client.stream_pricing(instruments):
            price = data.get('PRICE', None)
            if price:
                print(price.series())


loop = asyncio.get_event_loop()
loop.run_until_complete(stream('AUD_USD'))



import asyncio

from async_v20 import OandaClient
from async_v20.definitions.types import MarketOrderRequest

# Create an order to place
order = MarketOrderRequest('AUD_USD', 1)


async def account():
    async with OandaClient() as client:
        return await client.create_order(order=order)


loop = asyncio.get_event_loop()
response = loop.run_until_complete(account())

print(response)
